/*
 * xp_server - a simple mach external memory manager
 *
 * This server manages one file of records of a fixed size, and 
 * provides a memory object for it.
 *
 * HISTORY
 * Edit on Sep 30 1990 by Bill Schilit
 *	created based on Karin Peterson's external pager of 7/2/90.
 * Hacked by Erez...  Spring 1992.
 * e-6118y.
 */

#include <mach_interface.h>
#include <mach/mach_param.h>
#include <mach/message.h>
#include <mach/notify.h>
#include <sys/types.h>
#include <sys/file.h>
#include <sys/wait.h>
#include <servers/netname.h>
#include <mach.h>
#include <mig_errors.h>
#include <cthreads.h>
#include <signal.h>
#include <stdio.h>
#include <errno.h>
#include <assert.h>
#include <libc.h>

#include "xp.h"

extern int errno;

				/* the file pointer and addresses of objects */
FILE *fp;
int filesize;
int def_page_size;
int twopages;
int clients = 0;

port_t
  server_port = PORT_NULL,	/* server recvs msgs here */
  kernel_port = PORT_NULL;	/* kernel control port */

boolean_t 
  pending_completes = FALSE,
  at_end = FALSE,
  no_pager = FALSE,
  xp_pager_debug = TRUE;

typedef struct msg {
  msg_header_t	header;
  char 		data[MSG_SIZE_MAX - sizeof(msg_header_t)];
} msg_t;

typedef struct reply {
  msg_header_t	header;
  msg_type_t	return_code_type;
  msg_return_t	return_code;
} reply_t;

#define MAY_CACHE	FALSE	/* caching or not? */
#define MO_CP_STRATEGY 	MEMORY_OBJECT_COPY_NONE	/* copy strategy */

extern boolean_t memory_object_server(msg_t *, msg_t *);

#ifndef __HC__
kern_return_t provide_data(memory_object_control_t, vm_offset_t,
			   vm_size_t, vm_prot_t);
#endif /* !__HC__ */


/*
 * panic() - a truly unruly error occured
 */
void panic(s)
     char *s;
{ fprintf(stdout,"panic: %s\n", s);
 exit(1);
}


/*
 * server_loop()
 *
 * Receives message requests for objects and call appropriate routines 
 * via mig interface memory_object_server().
 */
void server_loop()
{
  msg_t in, out;
  kern_return_t ret;

  fprintf(stdout,"server_loop: starting\n");
  
  for (;;) {
    in.header.msg_local_port = server_port;
    in.header.msg_size = sizeof(msg_t);

    fprintf(stdout,"server_loop: about to receive msg\n");
    fprintf(stdout,"server_loop: number of clients is %d\n", clients);
    
    ret = msg_receive(&in.header, RCV_INTERRUPT,0);
    if (ret != RCV_SUCCESS) {
      fprintf(stdout,"server_loop: receive failed %d\n",ret);
      port_deallocate(task_self(),server_port);
      exit(1);
    }

    fprintf(stdout,"server_loop: after receive\n");

    printf("server_loop: before ports %d %d %d %d\n",
	   in.header.msg_local_port, in.header.msg_remote_port,
	   out.header.msg_local_port, out.header.msg_remote_port);
    if ((memory_object_server(&in,&out)) &&
	(out.header.msg_remote_port != PORT_NULL)) {
      printf("server_loop: inside ports %d %d %d %d\n",
	     in.header.msg_local_port, in.header.msg_remote_port,
	     out.header.msg_local_port, out.header.msg_remote_port);
      msg_send(&out.header, SEND_INTERRUPT, 0);
    }
    printf("server_loop: after ports %d %d %d %d\n",
	   in.header.msg_local_port, in.header.msg_remote_port,
	   out.header.msg_local_port, out.header.msg_remote_port);

    fprintf(stdout,"server_loop: at end of loop (sleeping)\n\n");
    sleep(1);
  }
}


boolean_t verify_request(control, length, offset, msg)
     port_t control;
     vm_size_t length;
     vm_offset_t offset;
     char *msg;
{
  boolean_t ret;

  if (ret = (control == PORT_NULL))
    fprintf(stdout,"%s: control port is NULL\n",msg);
  else if (ret = (length % def_page_size))
    fprintf(stdout,"%s: length not multiple of pagesize\n",msg);
  else if (ret = (offset % def_page_size))
    fprintf(stdout,"%s: offset not page aligned\n",msg);

  return(ret);
}


/*
 * memory_object_init() - initialize paging object
 *
 * Notification that a task wants to map the given memory object
 * into a virtual address space.  The memory manager may issue
 * cache management requests on the control port, and will use
 * the object_name port to identify this data region.
 *
 */
kern_return_t memory_object_init(object, control, object_name, page_size)
     memory_object_t		object;
     memory_object_control_t	control;
     memory_object_name_t	object_name;
     vm_size_t			page_size;
{
  if (at_end)
    return(KERN_FAILURE);

  if (xp_pager_debug)
    printf("(xp_pager) init: pager=%d, control=%d, name=%d\n",
	   object,control,object_name);

  assert(object != PORT_NULL);
  assert(control != PORT_NULL);
  assert(object_name != PORT_NULL);

  /* this should really be != PAGE_SIZE but sometimes its not! */

  if (page_size % def_page_size)
    panic("(xp_server) init: wrong page size");

  kernel_port = control;
  clients++;

  puts("init: before TRUE");
  if (memory_object_set_attributes(control, TRUE,
				   MAY_CACHE, MO_CP_STRATEGY) != KERN_SUCCESS)
    panic("(xp_server) init: 1 set_attributes");
  puts("init: after TRUE");
  return KERN_SUCCESS;
}


/*
 * memory_object_data_request() - request for data from paging object
 *
 * A request for data with at least the access specified has been made.
 * We try to return at least the specified data with as much access
 * as it can allow using memory_object_data_provided.  If a hardware
 * error has occurred then we may use memory_object_data_error.
 * 
 * If any pages have not been previously written then we reply with
 * memory_object_data_unavailable.
 *
 */
kern_return_t memory_object_data_request(object, control, offset,
					length, desired_access)
     memory_object_t		object;
     memory_object_control_t	control;
     vm_offset_t 		offset;
     vm_size_t 			length;
     vm_prot_t 			desired_access;
{
  if (at_end)
    return(KERN_FAILURE);

  if (xp_pager_debug)
    printf("(xp_server) data_request: object=%d, offset=0x%x, len=0x%x\n",
	   object, offset, length);
  
  if (length % def_page_size)
    panic("(xp_server) data_request: wrong page size");

  return(provide_data(control, offset, length, desired_access));
}


kern_return_t memory_object_data_unlock(object, control, offset, 
					length, desired_access)
     memory_object_t		object;
     memory_object_control_t	control;
     vm_offset_t		offset;
     vm_size_t 			length;
     vm_prot_t 			desired_access;
{
  vm_prot_t disallow_access;

  if (at_end)
    return(KERN_FAILURE);

  if (xp_pager_debug)
    printf("%s: obj=%d, ctrl=0x%x, offset=0x%x, len=0x%x, access=0x%x\n",
	   "(xp_server) data_unlock", object, control, offset,
	   length, desired_access);

				/* decide on proper access */
  disallow_access = VM_PROT_ALL;
  if (desired_access & VM_PROT_WRITE)
    disallow_access &= ~VM_PROT_WRITE;
  if (desired_access & VM_PROT_READ)
    disallow_access &= ~VM_PROT_READ;
  if (desired_access & VM_PROT_EXECUTE)
    disallow_access &= ~VM_PROT_EXECUTE;
  if (memory_object_lock_request(control, offset, length,
				 TRUE, TRUE, disallow_access,
 				 object) != KERN_SUCCESS)
    panic("(xp_server) data_unlock: lock_request failed");

  return(KERN_SUCCESS);
}


kern_return_t memory_object_data_write(object, control, offset, 
				       addr, data_count)
     memory_object_t		object;
     memory_object_control_t	control;
     vm_offset_t		offset;
     pointer_t			addr;
     vm_size_t			data_count;
{
  void write_pages();

  if (xp_pager_debug)
    printf("(xp_server) data_write: object=%d, offset=0x%x, len=0x%x\n",
	   object, offset, data_count);

  if (verify_request(control, offset, data_count, "memory_object_data_write"))
    return(KERN_FAILURE);

  write_pages(control, offset, addr, data_count);
  vm_deallocate(task_self(), addr, data_count);
  
  return(KERN_SUCCESS);
}


kern_return_t memory_object_lock_completed(object, control, offset, length)
     memory_object_t		object;
     memory_object_control_t	control;
     vm_offset_t		offset;
     vm_size_t			length;
{
  if (xp_pager_debug)
    printf("%s object=%d, ctrl=0x%x, offset=0x%x, len=0x%x\n",
	   "(xp_server) lock_completed:",
	   object, control, offset, length);

  if (verify_request(control, offset, length, "memory_object_lock_completed"))
    return(KERN_FAILURE);

  pending_completes = FALSE;
  
  if (at_end) {
    fflush(fp); 
    fclose(fp);
    if (xp_pager_debug)
      printf("lock_completed: exiting the server\n");
    exit(0);
  }
  return(KERN_SUCCESS);
}


kern_return_t memory_object_terminate(object, control, object_name)
     memory_object_t		object;
     memory_object_control_t	control;
     memory_object_name_t	object_name;
{
  if (xp_pager_debug)
    printf("(xp_server) object_terminate: object=%d, control=%d, name=%d\n",
	   object, control, object_name);

  fflush(fp);
  clients--;
  return(KERN_SUCCESS);
}


kern_return_t
memory_object_copy(old_object, old_control, offset, length, new_object)
     memory_object_t		old_object;
     memory_object_control_t	old_control;
     vm_offset_t		offset;
     vm_size_t			length;
     memory_object_t		new_object;
{ panic("(xp_server) object_copy: called");
  return(KERN_FAILURE);
}


/*
 * set all protection rights and provide data if needed
 */
kern_return_t provide_data(control, offset, length, desired_access)
     memory_object_control_t	control;
     vm_offset_t 		offset;
     vm_size_t 			length;
     vm_prot_t 			desired_access;
{
  vm_prot_t disallow_access;
  kern_return_t ret;
  char *buffer;

				/* give exact access */
  disallow_access = VM_PROT_ALL;
  if (desired_access & VM_PROT_WRITE)
    disallow_access &= ~VM_PROT_WRITE;
  if (desired_access & VM_PROT_READ)
    disallow_access &= ~VM_PROT_READ;
  if (desired_access & VM_PROT_EXECUTE)
    disallow_access &= ~VM_PROT_EXECUTE;

  if (xp_pager_debug)
    printf("%s: ctrl=0x%x, offset=0x%x, len=0x%x, disallow_access=0x%x\n",
	   "(xp_server) provide_data",control,offset,length,disallow_access);

  if (offset < 0 || offset > filesize || offset+length-1 > filesize) {
    fprintf(stdout,"provide_data: offset out of bounds\n");
    memory_object_data_error(control,offset,length,KERN_MEMORY_ERROR);
    return(KERN_FAILURE);
  }

  if (!(buffer = (char *) malloc(length))) {
    fprintf(stdout,"provide_data: no memory for buffer\n");
    memory_object_data_error(control,offset,length,KERN_MEMORY_FAILURE);
    return(KERN_FAILURE);
  }

  if (fseek(fp,offset,0) < 0) {
    free(buffer);
    fprintf(stdout,"provide_data: seek to offset in file failed\n");
    memory_object_data_error(control,offset,length,KERN_MEMORY_FAILURE);
    return(KERN_FAILURE);
  }

  if (!fread(buffer,1,length,fp)) {
    free(buffer);
    perror("provide_data can't read from file");
    memory_object_data_error(control,offset,length,KERN_MEMORY_FAILURE);
    return(KERN_FAILURE);
  }

  ret = memory_object_data_provided(control, offset,
				    (pointer_t) buffer,
				    length, disallow_access);
  free(buffer);
  return(ret);
}


/*
 * write_pages
 *
 * Write modified pages back to the file
 *
 */
void write_pages(memory_control, offset, data, data_count)
     port_t memory_control;
     vm_offset_t offset;
     char *data;
     int data_count;
{
  fprintf(stdout,"write_pages: start\n");

  if (offset < 0 || offset > filesize) {
    fprintf(stdout,"write_pages: offset out of bounds\n");
    return;
  }

  if (fseek(fp,offset,0) < 0) {
    fprintf(stdout,"write_pages: fseek failed\n");
    return;
  }

  if (!fwrite(data,data_count,1,fp)) {
    fprintf(stdout,"write_pages: fwrite failed\n");
    return;
  }

  memory_object_lock_request(memory_control,offset,data_count,
			     TRUE,TRUE,VM_PROT_ALL, PORT_NULL);
/* 			     FALSE,TRUE,VM_PROT_ALL, PORT_NULL); */
  vm_deallocate(task_self(), (vm_address_t) data, (vm_size_t) data_count);
  
  fprintf(stdout,"write_pages: exit\n");
}


void init_file_data_struct()
{
  fprintf(stdout,"init_file_data_struct: enter\n");
  
  if (!(fp = fopen("testfile","r+"))) {
    fprintf(stdout,"init_file_data_struct: can't open file\n");
    exit(1);
  }

  def_page_size = getpagesize();
  printf("server: page size = %d\n",def_page_size);

  twopages = 2 * def_page_size;
  
  filesize = twopages;

  fprintf(stdout,"init_file_data_struct: exit\n");
}


/*
 * register_server
 *
 * Register server with netname service.
 *
 */
void register_server()
{
  port_t port = PORT_NULL;

  while (port == PORT_NULL) {
    if (netname_look_up(name_server_port,"","NM_ACTIVE",&port) != KERN_SUCCESS)
      fprintf(stdout,"register_server: waiting for name server...\n");
    sleep(2);			/* doesn't work in threads */
  }

  port_deallocate(task_self(),port);

  if (netname_look_up(name_server_port, "", XP_SERVER_NAME, &port)
      == KERN_SUCCESS) {
    fprintf(stdout, "register_server: %s is already registered\n",
	    XP_SERVER_NAME);
    exit(1);
  }

  if (port_allocate(task_self(),&server_port) != KERN_SUCCESS) {
    fprintf(stdout,"register_server: can't allocate server port\n");
    exit(1);
  }

  if (netname_check_in(name_server_port, XP_SERVER_NAME,
		       task_self(),server_port) != NETNAME_SUCCESS) {
    fprintf(stdout,"register_server: can't check in server name\n");
    exit(1);
  }

  fprintf(stdout,"register_server: server name %s registered\n",
	  XP_SERVER_NAME);
}


void unregister_server()
{ if (netname_check_out(name_server_port,
			XP_SERVER_NAME, task_self()) != NETNAME_SUCCESS) {
    fprintf(stdout,"unregister_server: check out failed\n");
    exit(1);
  }
}


int cleanup_and_exit(sig,code,scp,addr)
     int sig, code;
     struct sigcontext *scp;
     char *addr;
{
  fprintf(stdout,"cleanup_and_exit: clients = %d\n",clients);
  
  unregister_server();
  pending_completes = TRUE;
  at_end = TRUE;
  if (memory_object_lock_request(kernel_port, 0, twopages, TRUE, TRUE, 
				 VM_PROT_ALL, server_port) != KERN_SUCCESS) {
    fprintf(stdout,"cleanup_and_exit: lock request failed\n");
    exit(1);
  }
  if (!MAY_CACHE && (clients == 0)) { /* nothing to do */
    fflush(fp);
    fclose(fp);
    fprintf(stdout,"cleanup_and_exit: exiting\n");
  }

  fprintf(stdout,"cleanup_and_exit: object shutting down\n");
  return 0;
}


void main(argc, argv)
     int argc;
     char *argv[];
{

  signal(SIGHUP, cleanup_and_exit);
  signal(SIGINT, cleanup_and_exit);
  
  cthread_init();
  register_server();
  init_file_data_struct();
  no_pager = FALSE;
  server_loop();
				/* should not reach here */
  fprintf(stdout,"main: exiting?\n");
  exit(1);
}
