/* Copyright (C) 1988, 1989, 1992 Free Software Foundation, Inc.
This file is part of GNU Make.

GNU Make is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 1, or (at your option)
any later version.

GNU Make is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with GNU Make; see the file COPYING.  If not, write to
the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.  */

#include <sys/wait.h>

#include "make.h"
#include "commands.h"

#include "pvm3.h"
#include "pgmake.h"

static char rcsid[]=
"$Id: remote-pvm3.c,v 1.9 1993/05/13 16:21:05 lih Exp lih $";

char *remote_description = "pvm";

#define MAXHOSTS    256

#define MAKEDAEMON  "pvm3_ad"
#define MAKEARCH    "SUN4"

/* Normalized path name of the current directory.  */
static char *normalized_cwd;

static int              nhost;
static struct hostinfo *hostp;	/* Points to static table */
static int             *tids;	/* Points to task ids */

/* determine if pvm has been initialized */
static int pvm_inited = 0;

extern int remote_jobs_flag;
extern int pvm_debug_flag;


/* Return nonzero if the next job should be done remotely.  */
int
normalize_directory(char *wd)
{
  char buf[BUFSIZ];
  char *cp;

  strcpy (buf, wd);

  cp = &buf[0];
  cp = index (cp, '/');		/* Picks up initial slash */
  cp++;
  cp = index (cp, '/');		/* Picks up next slash */
  cp++;
  cp = index (cp, '/');		/* Picks up first slash in /u/zing */

  strcpy (wd, cp);
}


int
start_remote_job_p ()
{

  if (remote_jobs_flag == 0 && getenv("GNU_MAKE_USE_PVM") == 0) {
    if (pvm_debug_flag)
      fprintf(stderr,"make: start_remote_job_p, using vanilla local jobs\n");
    return 0;
  }

  if (!pvm_inited) {
    
    int mytid, status, narch;
    
    /* PVM Startup */
    mytid  = pvm_mytid();
    status = pvm_config(&nhost, &narch, &hostp);
    
    if (status) {
      perror("pvm_config");
      return 0;
    }

    /* Setup data structures */
    hostp = (struct hostinfo *) malloc (sizeof (struct hostinfo) * MAXHOSTS);
    tids  = (int *) malloc (sizeof (int) * MAXHOSTS);
    {
      int i;
      for (i=0; i<MAXHOSTS; i++)
	tids[i]=-1;		/* Initialize */
    }

    normalized_cwd = (char *) xmalloc (GET_PATH_MAX);
    strcpy (normalized_cwd, starting_directory);
    normalize_directory (normalized_cwd); /* Normalizes in place */

    if (pvm_debug_flag) {
      fprintf(stderr,"make: start_remote_job_p, using pvm!\n");
      fprintf(stderr,"make: start_remote_job_p, mytid is 0x%x\n", mytid);
    }
    pvm_inited++;
  }
  
  return 1;
}


/* Start a remote job running the command in ARGV,
   with environment from ENVP.  It gets standard input from STDIN_FD.  On
   failure, return nonzero.  On success, return zero, and set *USED_STDIN
   to nonzero if it will actually use STDIN_FD, zero if not, set *ID_PTR to
   a unique identification, and set *IS_REMOTE to zero if the job is local,
   nonzero if it is remote (meaning *ID_PTR is a process ID).  */

int
start_remote_job (argv, envp, stdin_fd, is_remote, id_ptr, used_stdin)
     char **argv, **envp;
     int stdin_fd;
     int *is_remote;
     int *id_ptr;
     int *used_stdin;
{
  int  sbuf;			/* PVM send buffer */
  int  i, n;			/* Loop control variables */

  if (pvm_debug_flag)
    fprintf(stderr,"start_remote_job: *** ENTERED\n");

#ifdef notdef
  if (!stdin_fd) {
    if (pvm_debug_flag)
      fprintf(stderr,"start_remote_job: stdin_fd = 0, not doing anything\n");
    return -1;
  }
#endif

  sbuf = pvm_mkbuf(PvmDataDefault); /* Create a Default encoding buf */
  (void) pvm_setsbuf(sbuf);         /* Set to be current buffer      */

  /*****
   *****  Pack debugging flag
   *****/
  pvm_pkint(&pvm_debug_flag, 1, 1);
  
  /*****
   *****  Pack working directory
   *****/
  if (pvm_debug_flag)
    fprintf(stderr,"start_remote_job: pack wd %s\n",normalized_cwd);
  pvm_pkstr(normalized_cwd);

  /*****
   *****  Pack argv
   *****/
  if (argv)
    for (n = 0; argv[n]; n++);
  else
    n = 0;
  pvm_pkint(&n, 1, 1);		         /* Pack the length of vector */
  for (i = 0; i < n; i++)	         /* Pack each string */
    pvm_pkstr(argv[i]);

  /*****
   *****  Pack envp
   *****/
  if (envp)
    for (n = 0; envp[n]; n++);
  else
    n = 0;
  pvm_pkint(&n, 1, 1);		         /* Pack the length of vector */
  for (i = 0; i < n; i++)	         /* Pack each string */
    pvm_pkstr(envp[i]);

  {
    int index;
    int result = -1;
    int retry  = -1;
    for (index=0; tids[index]>-1; index++);

    if (pvm_debug_flag)
      pvm_serror(1);

    /*****  SPAWN the job, retrying if necessary.  This is mainly because
     *****    it is possible in PVM to have deadbeat nodes that are alive
     *****    but unable to execute remote commands.
     *****/
    while (result<=0 && retry<PGMAKE_SPAWN_RETRY_LIMIT) {
      result = pvm_spawn(MAKEDAEMON, (char**)0,
			 PvmTaskArch, MAKEARCH, 1, &tids[index]);
      if (result <= 0)
	if (pvm_debug_flag) {
	  fprintf(stderr,"pvm_spawn: error %d\n",result);
	  fprintf(stderr,"pvm_spawn: tid was (hex)%x\n",tids[index]);
	  fprintf(stderr,"pvm_spawn: tid was (dec)%d\n",tids[index]);
	  if (retry<PGMAKE_SPAWN_RETRY_LIMIT)
	    fprintf(stderr,"pvm_spawn: retrying...\n");
	}

      retry++;
    }

    if (retry>=PGMAKE_SPAWN_RETRY_LIMIT) {
      if (pvm_debug_flag)
	fprintf(stderr,"pvm_spawn: retry error, tried %d times\n",retry);
    } else {
      if (pvm_debug_flag)
	fprintf(stderr,"pvm_spawn: needed %d retries\n",retry);
    }

    if (result <= 0) {
      if (pvm_debug_flag) {
	fprintf(stderr,"pvm_spawn: error %d\n",result);
	fprintf(stderr,"pvm_spawn: tid was (hex)%x\n",tids[index]);
	fprintf(stderr,"pvm_spawn: tid was (dec)%d\n",tids[index]);
      }
      tids[index] = -1;
      return -1;
    }

    if (pvm_debug_flag) {
      fprintf(stderr,"start_remote_job: started tid=%x, index=%d\n",
	      tids[index], index);
      fprintf(stderr,"start_remote_job: send packet to tid=%x\n",
	      tids[index]);
    }

    pvm_send( tids[index], PGMAKE_JOB_BEGIN );

    *id_ptr = tids[index];
    *is_remote = 1;
    *used_stdin = 0;
  }

  if (pvm_debug_flag)
    fprintf(stderr,"start_remote_job: *** LEAVING\n");

  return 0;
}


/* Get the status of a dead remote child.  Block waiting for one to die
   if BLOCK is nonzero.  Set *EXIT_CODE_PTR to the exit status, *SIGNAL_PTR
   to the termination signal or zero if it exited normally, and *COREDUMP_PTR
   nonzero if it dumped core.  Return the ID of the child that died,
   0 if we would have to block and !BLOCK, or < 0 if there were none.  */

int
remote_status (exit_code_ptr, signal_ptr, coredump_ptr, block)
     int *exit_code_ptr, *signal_ptr, *coredump_ptr;
     int block;
{
  int  msgbuf;
  int  tid, status, i;
  char stdout_buf[BUFSIZ];
  char stderr_buf[BUFSIZ];

  if (!pvm_inited)		/* do nothing if PVM was not initialized */
    return -1;

  if (pvm_debug_flag)
    fprintf(stderr,"remote_status: called with block=%d\n", block);

  if (block)
    if (pvm_debug_flag)
      fprintf(stderr,"remote_status: make wants us to block\n");
  
  /* Receive anything */
  while (1) {

    if (msgbuf = pvm_nrecv (-1, PGMAKE_STDOUT)) {
      if (pvm_debug_flag)
	fprintf(stderr,"remote_status: stdout message\n");
      pvm_upkint (&tid, 1, 1);	/* Unpack the id     */
      pvm_upkstr (stdout_buf);	/* Unpack the stdout */
      fprintf(stdout,"[%x] stdout: %s\n", stdout_buf);
    }
      
    if (msgbuf = pvm_nrecv (-1, PGMAKE_STDERR)) {
      if (pvm_debug_flag)
	fprintf(stderr,"remote_status: stderr message\n");
      pvm_upkint (&tid, 1, 1);	/* Unpack the id     */
      pvm_upkstr (stderr_buf);	/* Unpack the stderr */
      fprintf(stderr,"[%x] stderr: %s\n", stderr_buf);
    }

    if (msgbuf = pvm_nrecv (-1, PGMAKE_JOB_END))
      break;
  }

  if (pvm_debug_flag)
    fprintf(stderr,"remote_status: message received!\n");
  
  pvm_upkint (&tid, 1, 1);	/* Unpack the id     */
  pvm_upkint (&status, 1, 1);	/* Unpack the status */

#ifdef FUZCOMMENT
  pvm_upkstr (stdout_buf);	/* Unpack the stdout */
  pvm_upkstr (stderr_buf);	/* Unpack the stderr */
#endif
  
  for (i=0; i<MAXHOSTS; i++)
    if (tid==tids[i])
      break;

  if (i==MAXHOSTS) {
    if (pvm_debug_flag)
      fprintf(stderr,"remote_status: couldn't find tid\n");
    return -1;
  } else {
    if (pvm_debug_flag) {
      fprintf(stderr,"remote_status: tid=%x\n",    tid);
      fprintf(stderr,"remote_status: status=%d\n", status);
    }

#ifdef FUZCOMMENT
    if (pvm_debug_flag)
      fprintf(stderr,"remote_status: stdout:\n");
    fprintf(stdout,"%s", stdout_buf); /* print stdout */
    if (pvm_debug_flag)
      fprintf(stderr,"remote_status: stderr:\n");
    fprintf(stderr,"%s", stderr_buf); /* print stderr */
#endif 
    
    /* decode return status from wait() */
    *exit_code_ptr = 0;
    *signal_ptr    = 0;
    *coredump_ptr  = 0;

    if (WIFEXITED(status)) {
      *exit_code_ptr = WEXITSTATUS(status);
    } else if (WIFSIGNALED(status)) {
      *signal_ptr = WTERMSIG(status);
      *coredump_ptr = status & 0x80; /* was a core dump produced? */
    }
    
    tids[i] = -1;
    return tid;
  }

}


/* Block asynchronous notification of remote child death.
   If this notification is done by raising the child termination
   signal, do not block that signal.  */
void
block_remote_children ()
{
  return;
}


/* Restore asynchronous notification of remote child death.
   If this is done by raising the child termination signal,
   do not unblock that signal.  */
void
unblock_remote_children ()
{
  return;
}


/* Send signal SIG to child ID.  Return 0 if successful, -1 if not.  */
int
remote_kill (tid, sig)
     int tid;
     int sig;
{
  int sbuf;
  int oldbuf;

  if (pvm_debug_flag)
    fprintf(stderr,"remote_kill: gmake sent %x, signal %d\n",tid,sig);

  sbuf   = pvm_mkbuf(PvmDataDefault); /* Create a Default encoding buf */
  oldbuf = pvm_setsbuf(sbuf);         /* Set to be current buffer      */

  pvm_pkint(&sig, 1, 1);
  pvm_send (tid, PGMAKE_KILL);

  (void) pvm_setsbuf(oldbuf);
  pvm_freebuf(sbuf);

  return -1;
}
