/*-
 * signal.c --
 *	Signal number and name mapping
 *
 * Copyright (c) 1982 Regents of the University of California.
 * Copyright (c) 1992 by Andreas Stolcke
 *
 * Permission to use, copy, modify, and distribute this
 * software and its documentation for any non-commercial purpose
 * and without fee is hereby granted, provided that the above copyright
 * notice appears in all copies.  The University of California,
 * and Andreas Stolcke make no representations about
 * the suitability of this software for any purpose.  It is provided
 * "as is" without express or implied warranty.
 *
 */
#ifndef lint
static char *rcsid =
"$Id: signal.c,v 1.1 1993/03/11 06:32:48 stolcke Exp $ ICSI (Berkeley)";
#endif /* not lint */

#include    <strings.h>
#include    <sys/signal.h>

#include    "customs.h"

static struct signal {
    int	signo;			/* OS specific signal number */
    char *name;			/* short signal name */
} signalList[] = {
    { 0,	"" },		/* Probe for process */
    { SIGHUP,	"HUP" },	/* hangup */
    { SIGINT,	"INT" },	/* interrupt */
    { SIGQUIT,	"QUIT" },	/* quit */
    { SIGILL,	"ILL" },	/* illegal instruction */
    { SIGTRAP,	"TRAP" },	/* trace trap */
#ifdef SIGIOT
    { SIGIOT,	"IOT" },	/* IOT instruction */
#else
    { SIGABRT,	"ABRT" },	/* Process abort signal */
#endif
    { SIGEMT,	"EMT" },	/* EMT instruction */
    { SIGFPE,	"FPE" },	/* floating point exception */
    { SIGKILL,	"KILL" },	/* kill (cannot be caught or ignored) */
    { SIGBUS,	"BUS" },	/* bus error */
    { SIGSEGV,	"SEGV" },	/* segmentation violation */
    { SIGSYS,	"SYS" },	/* bad argument to system call */
    { SIGPIPE,	"PIPE" },	/* write on a pipe with no one to read it */
    { SIGALRM,	"ALRM" },	/* alarm clock */
    { SIGTERM,	"TERM" },	/* software termination signal from kill */
    { SIGURG,	"URG" },	/* urgent condition on IO channel */
    { SIGSTOP,	"STOP" },	/* sendable stop signal not from tty */
    { SIGTSTP,	"TSTP" },	/* stop signal from tty */
    { SIGCONT,	"CONT" },	/* continue a stopped process */
#ifdef SYSV
    { SIGCLD,	"CLD" },	/* to parent on child stop or exit */
#else
    { SIGCHLD,	"CHLD" },	/* to parent on child stop or exit */
#endif
    { SIGTTIN,	"TTIN" },	/* to readers pgrp upon background tty read */
    { SIGTTOU,	"TTOU" },	/* like TTIN for output with LTOSTOP flag */
#ifdef SYSV
    { SIGPOLL,	"POLL" },	/* input/output possible signal */
#else
    { SIGIO,	"IO" },		/* input/output possible signal */
#endif
#ifdef SIGXCPU
    { SIGXCPU,	"XCPU" },	/* exceeded CPU time limit */
#else
    { 99,	"" },		/* undefined */
#endif
#ifdef SIGXFSZ
    { SIGXFSZ,	"XFSZ" },	/* exceeded file size limit */
#else
    { 99,	"" },		/* undefined */
#endif
    { SIGVTALRM,"VTALRM" },	/* virtual time alarm */
    { SIGPROF,	"PROF" },	/* profiling time alarm */
    { SIGWINCH, "WINCH" },	/* window changed */
#ifdef SIGLOST
    { SIGLOST,	"LOST" },	/* resource lost (eg, record-lock lost) */
#else
    { 99,	"" },		/* undefined */
#endif
    { SIGUSR1,	"USR1" },	/* user defined signal 1 */
    { SIGUSR2,	"USR2" },	/* user defined signal 2 */
    { 99,	"invalid" }	/* Undefined on this system */
};

/*-
 *-----------------------------------------------------------------------
 * Signal_ToHost --
 *	Map an SIGNAL_ value to host specific signal number
 *
 * Results:
 *	The host specific value.
 *
 * Side Effects:
 *	None.
 *
 *-----------------------------------------------------------------------
 */
int
Signal_ToHost (signal)
    int	    	  	signal;
{
    if (signal < 0 || signal > SIGNAL_INVAL) {
	signal = SIGNAL_INVAL;
    }
    return (signalList[signal].signo);
}

/*-
 *-----------------------------------------------------------------------
 * Signal_FromHost --
 *	Map a host specific signal number to a SIGNAL_ value.
 *
 * Results:
 *	The SIGNAL_ value.
 *
 * Side Effects:
 *	None.
 *
 *-----------------------------------------------------------------------
 */
int
Signal_FromHost (signo)
    int	    	 signo;
{
    int signal;

    for (signal = 0; signal < SIGNAL_INVAL; signal++) {
	if (signalList[signal].signo == signo) {
	    break;
	}
    }
    return (signal);
}

/*-
 *-----------------------------------------------------------------------
 * Signal_ToName --
 *	Map a SIGNAL_ value to its name
 *
 * Results:
 *	The SIGNAL_ value.
 *
 * Side Effects:
 *	None.
 *
 *-----------------------------------------------------------------------
 */
char *
Signal_ToName (signal)
    int	    	 signal;
{
    if (signal < 0 && signal > SIGNAL_INVAL) {
	signal = SIGNAL_INVAL;
    }
    return (signalList[signal].name);
}

/*-
 *-----------------------------------------------------------------------
 * Signal_FromName --
 *	Map a signal name to a SIGNAL_ value.
 *
 * Results:
 *	The SIGNAL_ value.
 *
 * Side Effects:
 *	None.
 *
 *-----------------------------------------------------------------------
 */
int
Signal_FromName (name)
    char	*name;
{
    int signal;

    for (signal = 0; signal < SIGNAL_INVAL; signal++) {
	if (strcmp(signalList[signal].name, name) == 0) {
	    break;
	}
    }
    return (signal);
}

