#include <sys/param.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/fcntl.h>
#include <sys/errno.h>
#include <sys/time.h>
#include <rpc/types.h>
#include <nfs/nfs.h>
#include <rpc/rpc.h>
#include <rpc/pmap_clnt.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <ndbm.h>

#include "../nfsmgrd/nfsmgr_common.h"
#include "checksumd.h"
#include "../md4/optimized/md4.h"


			/* DEFINITIONS */
#define OPTIONS "slIf:F:D:"		/* getopt options */
#define USAGE	"[-Ils] [-F dir] [-D lvl] [-f dbfile]"
#define DEFAULT_DIR "/usr/local" /* filesystem/directory to initialize */
#define DEFAULT_CKMFILE "md4chksums" /* checksums database file */


			/* EXTERNALS */
extern void pclose(), fclose();
extern char *optarg, *sys_siglist[];
extern int errno, perror(), lstat(), fprintf(), printf(), getopt();
extern int fread(), bcopy(), registerrpc(), bzero();


			/* LOCALS */
int initialize_db(), list_db(), update_db(), serve_db();
int *MDfile(), reinitialize_database();


			/* GLOABALS */
static DBM *db = (DBM *) NULL;


			/* DEFAULTS */
int opt_I = 0;			/* don't initialize the database */
int opt_D = 0;			/* no debugging */
int opt_l = 0;			/* list database */
int opt_s = 1;			/* serve database by default */
char *opt_F = (char *) NULL;	/* filesystem/directory to initialize */
char *opt_f = (char *) NULL;	/* checksums database file */


int main(argc,argv)
     int argc;
     char **argv;
{
  char c;
  int retval;

				/* parse options */
  while ((c = getopt(argc, argv, OPTIONS)) != EOF) {
    switch (c) {
    case 'I':			/* initialize database */
      opt_I = 1;
      break;
    case 'D':			/* debug level */
      opt_D = atoi(optarg);
      break;
    case 'F':			/* filesystem/dir to init */
      opt_F = optarg;
      break;
    case 'l':			/* list the database */
      opt_l = 1;
      break;
    case 's':			/* serve the database */
      opt_s = 1;
      break;
    case 'f':			/* database file name */
      opt_f = optarg;
      break;
    default:
      fprintf(stderr, "Usage: %s %s\n", argv[0], USAGE);
      exit(1);
    }
  }

			/* INITIALIZE VARIABLES */
  if (!opt_F)
    opt_F = DEFAULT_DIR;
  if (!opt_f)
    opt_f = DEFAULT_CKMFILE;

			/* PERFORM ACTIONS */
  if (opt_I) {			/* initialize */
    retval = initialize_db(opt_F);
    exit(retval);
  }
  if (opt_l) {			/* list database */
    retval = list_db(opt_F);
    exit(retval);
  }
  if (opt_s) {			/* serve database */
    retval = serve_db(opt_F);
    exit(retval);
  }

  exit(0);
}


/*
 * Get the checksum for a given pathname (string)
 * Returns 0 if not found, or error.
 */
int *sget_checksum(path)
     char **path;
{
  datum key, content;
  chksumdat_t cs;
  static int *ckm;

  int retval;
  static ino_t inode;
  struct stat statbuf;


				/* open database */
  if (db == (DBM *) NULL) {
    db = dbm_open(opt_f, O_RDWR|O_CREAT, 0777);
    if (db == (DBM *) NULL) {
      perror(opt_f);
      return (int *) NULL;
    }
  }

				/* initialize 4 integer array */
  if (ckm == (int *) NULL) {
    ckm = (int *) calloc(4, sizeof(int));
    if (ckm == (int *) NULL) {
      printf("no more memory\n");
      return (int *) NULL;
    }
  }
  bzero((char *)ckm, sizeof(int)*4);

				/* get inode first */
  if (lstat(*path, &statbuf) < 0) { /* stat path */
    perror(*path);
    return ckm;
  }
  if (!S_ISDIR(statbuf.st_mode) &&
      !S_ISREG(statbuf.st_mode) &&
      !S_ISLNK(statbuf.st_mode)) {
    fprintf(stderr, "skipping path \"%s\" -- not comparable\n", *path);
    return ckm;
  }
  inode = statbuf.st_ino;
  if (opt_D) {
    printf("\"%s\":\n", *path);
    printf("%ld %ld :", statbuf.st_ino, statbuf.st_mtime);
  }

/*
 * Now we have the inode.  See if key is in database.
 */
				/* build key */
  key.dsize = sizeof(ino_t);
  key.dptr = (char *)&inode;
  bzero((char *)&content, sizeof(datum));
  content = dbm_fetch(db, key);
/*
 * If did not find entry in db, or it was found but the modification time of
 * the curret inode was newer, then recompute the checksum and return it.
 */
  if (content.dptr == (char *)NULL) {
    if (opt_D) {
      printf("did not find data for key %ld\n", inode);
    }
    if ((ckm = MDfile(*path)) == (int *) NULL) {
      printf("cannot compute checksum\n");      
    } else {
      if (opt_D)
	printf("%08x|%08x|%08x|%08x\n", ckm[0], ckm[1], ckm[2], ckm[3]);
      retval = update_db(*path, statbuf.st_ino, statbuf.st_mtime, ckm);
      if (retval != 0)
	fprintf(stderr, "could not update database \"%s\"\n", opt_f);
    }
  } else {			/* it did find an entry in the database */
    /* find out the mtime of the entry  */
    bcopy(content.dptr, (char *) &cs, content.dsize);
    if (statbuf.st_mtime > cs.mtime) { /* then db entry is too old! */
      if (opt_D)
	printf("file in database is older -- recompute checksum\n");
      if ((ckm = MDfile(*path)) == (int *) NULL) {
	printf("cannot compute checksum\n");      
      } else {
	if (opt_D)
	  printf("%08x|%08x|%08x|%08x\n", ckm[0], ckm[1], ckm[2], ckm[3]);
	retval = update_db(*path, statbuf.st_ino, statbuf.st_mtime, ckm);
	if (retval != 0)
	  fprintf(stderr, "could not update database \"%s\"\n", opt_f);
      }
    } else {			/* then db entry is good */
      bcopy((char *) &cs.chksum_p[0], (char *)&ckm[0], sizeof(int)*4);
      if (opt_D) {
	printf("%ld %ld :", inode, cs.mtime);
	printf("%08x|%08x|%08x|%08x\n",
	       cs.chksum_p[0], cs.chksum_p[1], cs.chksum_p[2],
	       cs.chksum_p[3]);
      }
    }
  }
  if (dbm_error(db)) {
    fprintf(stderr, "unexpected error in database dbm_fetch\n");
    dbm_clearerr(db);
  }
  return ckm;
}


/*
 * Get the checksum for a given inode
 * Returns 0 if not found, or error.
 */
int *iget_checksum(inode)
     long *inode;
{
  datum key, content;
  chksumdat_t cs;
  static int ckm[4];
  static ino_t in;

				/* open database */
  if (db == (DBM *) NULL) {
    db = dbm_open(opt_f, O_RDWR|O_CREAT, 0777);
    if (db == (DBM *) NULL) {
      perror(opt_f);
      return (int *) NULL;
    }
  }
				/* build key */
  key.dsize = sizeof(ino_t);
  in = *inode;
  key.dptr = (char *)&in;
  bzero((char *)&content, sizeof(datum));
  content = dbm_fetch(db, key);
  if (content.dptr == (char *)NULL) {		/* did not find entry */
    if (opt_D) {
      printf("did not find data for key %ld\n", *inode);
    }
    bzero((char *)&ckm[0], sizeof(int)*4);
  } else {
    bcopy(content.dptr, (char *) &cs, content.dsize);
    bcopy((char *) &cs.chksum_p[0], (char *)&ckm[0], sizeof(int)*4);
    if (opt_D) {
      printf("%ld %ld :", *inode, cs.mtime);
      printf("%08x|%08x|%08x|%08x\n",
	     cs.chksum_p[0], cs.chksum_p[1], cs.chksum_p[2],
	     cs.chksum_p[3]);
    }
  }
  if (dbm_error(db)) {
    fprintf(stderr, "unexpected error in database dbm_fetch\n");
    dbm_clearerr(db);
  }
  return &ckm[0];
}


void cleanup_n_exit(signum)	/* clean up entries in port map. I don't */
     int signum;
{ char *signame;                /* want any mess left after I'm gone */

  if (opt_D)
    fprintf(stderr, "cleaning up and exiting\n");

  if (signum > NSIG)		/* make sure not going over max sig names */
    signame = "unknown";
  else
    signame = sys_siglist[signum];

  if (opt_D)
    fprintf(stderr, "Prog. %u, ver. %u, received signal %d (%s) ",
	    CHKSUMD_PROGRAM, CHKSUMD_VERSION, signum, signame);

  if (pmap_unset(CHKSUMD_PROGRAM, CHKSUMD_VERSION)) {
    if (opt_D)
      fprintf(stderr, "and was deleted\n");
  } else {
    if (opt_D)
      fprintf(stderr, "but could not be removed!!\n");
  }
				/* close/flush database */
  if (db) {
    dbm_close(db);
  }

  exit(signum);			/* it shouldn't have really ended */
}


/*
 * A signal hander so I can recompute the database
 */
int reinitialize_database()
{
  return initialize_db(opt_F);
}


/*
 * Serve a database.
 */
int serve_db(dir)
     char *dir;
{
  if (chdir(dir) < 0) {
    perror(dir);
    return errno;
  }

  if (registerrpc( CHKSUMD_PROGRAM, CHKSUMD_VERSION,
		   CHKSUMD_ichksum, iget_checksum,
		   xdr_long, xdr_checksum) != 0  ||
      registerrpc( CHKSUMD_PROGRAM, CHKSUMD_VERSION,
		  CHKSUMD_schksum, sget_checksum,
		  xdr_wrapstring, xdr_checksum) != 0) {
    fprintf(stderr, "could not serve databases with RPC\n");
    return 1;
  }
				/* set signal handlers */
  signal( SIGHUP, reinitialize_database);
  signal( SIGINT, cleanup_n_exit);
  signal( SIGQUIT, cleanup_n_exit);
  signal( SIGBUS, cleanup_n_exit);
  signal( SIGSEGV, cleanup_n_exit);
  signal( SIGALRM, cleanup_n_exit);
  signal( SIGTERM, cleanup_n_exit);
  signal( SIGTERM, cleanup_n_exit);

  if (opt_D)
    fprintf(stderr, "starting to serve database...\n");
  svc_run();

  return 0;
}


/*
 * List/dump contents of a database.
 */
int list_db(dir)
     char *dir;
{
  int retval;
  ino_t inode;
  datum key, content;
  chksumdat_t cs;

  if (chdir(dir) < 0) {
    perror(dir);
    return errno;
  }
				/* open database */
  if (db == (DBM *) NULL) {
    db = dbm_open(opt_f, O_RDWR|O_CREAT, 0777);
    if (db == (DBM *) NULL) {
      perror(opt_f);
      return errno;
    }
  }

				/* iterate over the db */
  key = dbm_firstkey(db);
  if ((retval = dbm_error(db))) {
    fprintf(stderr, "error in database dbm_firstkey\n");
    dbm_clearerr(db);
    return retval;
  }
  do {
    printf("%d %d %d %d\n",key.dptr[0],key.dptr[1],key.dptr[2],key.dptr[3]);
    content = dbm_fetch(db, key);
    if ((retval = dbm_error(db))) {
      fprintf(stderr, "unexpected error in database dbm_fetch\n");
      dbm_clearerr(db);
    }
    bcopy(key.dptr, (char *) &inode, key.dsize);
    bcopy(content.dptr, (char *) &cs, content.dsize);
    if (opt_D || opt_l) {
      printf("%ld %ld :", inode, cs.mtime);
      printf("%08x|%08x|%08x|%08x\n",
	     cs.chksum_p[0], cs.chksum_p[1], cs.chksum_p[2],
	     cs.chksum_p[3]);
    }
    key = dbm_nextkey(db);	/* advance to next key */
  } while (key.dptr != (char *) NULL);
  dbm_close(db);
  return 0;
}


/*
 * Initialize a new database.
 */
int initialize_db(dir)
     char *dir;
{
  FILE *FP;
  char path[MAX_PATH_SIZE];
  int len, *i4p, retval;
  struct stat statbuf;

  if (chdir(dir) < 0) {
    perror(dir);
    return errno;
  }
  if ((FP=popen("/bin/find . -print","r")) == (FILE *) NULL) {
    fprintf(stderr, "could not execute %s for read.\n", "/bin/find . -print");
    return errno;
  }
				/* first close database if open */
  if (db != (DBM *) NULL) {
    dbm_close(db);
  }
  db = dbm_open(opt_f, O_RDWR|O_CREAT, 0777); /* now reopen */
  if (db == (DBM *) NULL) {
    perror(opt_f);
    return errno;
  }

				/* process each file */
  while (fgets(path, MAX_PATH_SIZE, FP) != NULL) {
    len = strlen(path);		/* chop newline */
    path[len-1] = '\0';
    if (lstat(path, &statbuf) < 0) { /* stat path */
      perror(path);
      continue;
    }
    if (!S_ISDIR(statbuf.st_mode) &&
	!S_ISREG(statbuf.st_mode) &&
	!S_ISLNK(statbuf.st_mode)) {
      fprintf(stderr, "skipping path \"%s\" -- not comparable\n", path);
      continue;
    }
    if (opt_D) {
      printf("\"%s\":\n", path);
      printf("%ld %ld :", statbuf.st_ino, statbuf.st_mtime);
    }
    if ((i4p = MDfile(path)) == (int *) NULL) {
      printf("cannot compute checksum\n");      
    } else {
      if (opt_D)
	printf("%08x|%08x|%08x|%08x\n", i4p[0], i4p[1], i4p[2], i4p[3]);
      retval = update_db(path, statbuf.st_ino, statbuf.st_mtime, i4p);
      if (retval != 0) {
	fprintf(stderr, "could not update database \"%s\"\n", opt_f);
	return retval;
      }
    }
  }				/* end of "while (fgets" loop */
  pclose(FP);
  dbm_close(db);
  return 0;
}


/* MDfile(filename)
** Computes the message digest for a specified file.
** Prints out message digest, a space, the file name, and a carriage return.
** ACTUALLY RETURN POINTER TO int[4] buffer.  -Erez.
*/
int *MDfile(filename)
char *filename;
{
  FILE *f = fopen(filename,"rb");
  unsigned char X[64];
  static MDstruct MD;
  int b;

  if (f == NULL) {
    if (opt_D) {
      printf("%s can't be opened.\n", filename);
    }
    return (int *) NULL;
  }
  MDbegin(&MD);
  while ((b=fread(X,1,64,f))!=0)
    MDupdate(&MD,X,b*8);
  MDupdate(&MD,X,0);
/*   MDprint(&MD); */
/*   printf(" %s\n",filename); */
  fclose(f);
  return (int *) &MD.buffer[0];
}


/*
 * Update the database.
 */
int update_db(path, inode, mtime, ckm)
     char *path;
     ino_t inode;
     time_t mtime;
     int *ckm;
{
  datum key, content;
  chksumdat_t cs;
  int retval;

				/* open database */
  if (db == (DBM *) NULL) {
    db = dbm_open(opt_f, O_RDWR|O_CREAT, 0777);
    if (db == (DBM *) NULL) {
      perror(opt_f);
      return errno;
    }
  }

  key.dptr = (char *) &inode;
  key.dsize = sizeof(ino_t);
  cs.mtime = mtime;
  bcopy((char *)ckm, &cs.chksum_p[0], sizeof(int)*4);
  content.dptr = (char *) &cs;
  content.dsize = sizeof(chksumdat_t);
  dbm_store(db, key, content, DBM_REPLACE);
  if ((retval = dbm_error(db))) {
    fprintf(stderr, "error in database read/write\n");
    dbm_clearerr(db);
    return retval;
  }
  return 0;
}
